"""Utils."""

import contextlib
import os
import shlex
import subprocess
import tempfile

def get_contents(filename):
    """Get contents of given file."""
    with open(filename) as file:
        return file.read()

def sqlite_string(text):
    """Encode python string into sqlite string."""
    return "'{}'".format(text.replace("'", "''"))

def write_lines(filename, text):
    """Write text (iterable) to file."""
    with open(filename, "w") as file:
        for line in text:
            print(line, file=file)

@contextlib.contextmanager
def make_temporary_file(*args, **kwargs):
    """Temporary file context manager that returns filename."""
    _, filename = tempfile.mkstemp(*args, **kwargs)
    yield filename
    os.remove(filename)

def run_command(cmd, **kwargs):
    """Run command with environment variables in kwargs.

    Returns stdout and stderr output.
    """
    env = os.environ.copy()
    env.update(**kwargs)
    proc = subprocess.run(shlex.split(cmd), env=env, check=True,
                          capture_output=True)
    return proc.stdout, proc.stderr



"""Test page.py."""

from . import page

SQL = """
    PRAGMA foreign_keys=ON;
    INSERT INTO Files (filename) VALUES ('test.md');
    INSERT INTO Notes (id, title, filename) VALUES
        (0, '0', 'test.md'),
        (1, '1', 'test.md'),
        (2, '2', 'test.md');
    INSERT INTO Tags (id, tag) VALUES
        (0, '#test'),
        (1, '#test'),
        (2, '##test');
    INSERT INTO Bibliography (key, text) VALUES
        ('ref-test', 'Reference text.');
    INSERT INTO Citations (note, reference) VALUES
        (0, 'ref-test');
"""

def test_create_bibliography(mock_db):
    """Check create_bibliography output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_bibliography(conn)
    assert html == """<section id='references' title='References' class='level1'>
  <h1>
    References
  </h1>
  <dl>
    <dt>
      <a href='#ref-test'>
        [@test]
      </a>
    </dt>
    <dd>
      Reference text.
    </dd>
  </dl>
</section>"""

def test_create_tags(mock_db):
    """Check create_tags output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_tags(conn)
    assert html == """<section id='tags' title='Tags' class='level1'>
  <h1>
    Tags
  </h1>
  <ul>
    <li>
      <a href='###test'>
        ##test
      </a>
    </li>
    <li>
      <a href='##test'>
        #test
      </a>
    </li>
  </ul>
</section>"""

def test_create_tag_page(mock_db):
    """Check create_tag_page output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_tag_page(conn, "#test")
    assert html == """<section id='#test' title='#test' class='level1'>
  <h1>
    <a href='#tags' title='List of tags'>
      #test
    </a>
  </h1>
  <ul>
    <li>
      [0] 
      <a href='#0'>
        0
      </a>
    </li>
    <li>
      [1] 
      <a href='#1'>
        1
      </a>
    </li>
  </ul>
</section>"""

def test_create_reference_page(mock_db):
    """Check create_reference_page output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_reference_page(conn, "ref-test")
    assert html == """<section id='ref-test' title='ref-test' class='level1'>
  <h1>
    <a href='#references'>
      @test
    </a>
  </h1>
  <p>
    Reference text.
  </p>
  <ul>
    <li>
      [0] 
      <a href='#0'>
        0
      </a>
    </li>
  </ul>
</section>"""



"""Functions for mocking the database."""

import sqlite3
import pytest

from . import scan

@pytest.fixture
def mock_db():
    """Create an empty mock database with all the necessary tables."""
    with sqlite3.connect(":memory:") as conn:
        scan.initialize_database(conn)
        yield conn



"""Look for files that must be compiled."""

from itertools import chain, groupby
import fnmatch
import glob
import os
import shlex
import sys

from . import utils

def initialize_database(conn):
    """Initialize database with schema.sql."""
    filename = os.path.join(os.path.dirname(__file__), "schema.sql")
    sql = utils.get_contents(filename)
    conn.executescript(sql)
    conn.commit()

def fetch_files(conn):
    """Get files from the database."""
    return (row[0] for row in conn.execute("SELECT filename FROM Files"))

def is_recently_modified(timestamp):
    """Create function that checks if a file has been modified since the given
    timestamp.
    """
    return lambda f: os.path.exists(f) and os.path.getmtime(f) >= timestamp

def is_file_in_db(filename, conn):
    """Check if file is recorded in the database."""
    cur = conn.cursor()
    sql = "SELECT filename FROM Files WHERE filename = ?"
    for _ in cur.execute(sql, (filename,)):
        return True
    return False

def has_valid_pattern(filename, patterns):
    """Check if filename matches one of the patterns."""
    for pattern in patterns:
        if fnmatch.fnmatch(filename, pattern):
            return True
    return False

def files_in_path(path):
    """Recursively glob files in path."""
    if os.path.isfile(path):
        yield path
    elif os.path.isdir(path):
        basedir = os.path.join(path, "**")
        yield from filter(os.path.isfile, glob.iglob(basedir, recursive=True))

def files_in_paths(paths):
    """Recursively glob files in every path in paths."""
    new_files = set()
    for path in paths:
        new_files = new_files.union(files_in_path(path))
    return new_files

def find_new_files(conn, paths, patterns=("*.md",)):
    """Look for files that are not yet in the database."""
    condition = lambda x: os.path.isfile(x) and not is_file_in_db(x, conn) and \
            has_valid_pattern(x, patterns)
    return filter(condition, map(os.path.normpath, files_in_paths(paths)))

def input_files(conn, timestamp, paths, patterns=("*.md",)):
    """Generate files that must be rescanned/processed by pandoc.

    Neighbor notes of deleted/modified files don't have to be rescanned,
    since those notes don't get affected when notes they reference to are
    removed from the database.
    """
    modified = map(os.path.normpath, filter(is_recently_modified(timestamp),
                                            fetch_files(conn)))
    new = find_new_files(conn, paths, patterns)
    yield from filter(os.path.isfile, set(modified).union(new))

def delete_files_from_database(conn, files):
    """Delete files from the database."""
    cur = conn.cursor()
    cur.execute("PRAGMA foreign_keys=ON")
    for filename in files:
        cur.execute("DELETE FROM Files WHERE filename = ?", (filename,))
    conn.commit()

def run_script_on_database(conn, script):
    """Run script stored in file on database."""
    with open(script) as file:
        cur = conn.cursor()
        for line in file:
            cur.execute(line)
        conn.commit()

def group_by_file_extension(files):
    """Generate an iterator for each file extension.

    Each file with no file extension is given its own iterator.
    """
    def key(filename):
        root, ext = os.path.splitext(filename)
        return (ext, "") if ext else ("", root)
    groups = groupby(sorted(files, key=key), key=key)
    return map(lambda g: g[1], groups)

def build_command(files, output, options=""):
    """Construct a single pandoc command to run on input files.

    Return an empty string if there are no input files.
    """
    datadir = shlex.quote(os.path.abspath(os.path.dirname(__file__)))
    cmd = f"pandoc -Lzk.lua -Fpandoc-citeproc -Lrefs.lua --section-divs "\
            f"--data-dir {datadir} -Mlink-citations:true {options} "\
            "-o {} ".format(shlex.quote(output))
    empty = True
    for filename in files:
        empty = False
        cmd += " {}".format(shlex.quote(filename))
    return cmd if not empty else ""

def remove_outdated_files_from_database(conn, timestamp):
    """Remove outdated files from the database.

    This includes files that have been deleted from the file system, and files
    that have recently been modified.
    The second set of files will be added back to the database after scanning.
    """
    missing = filter(lambda x: not os.path.exists(x), fetch_files(conn))
    modified = filter(is_recently_modified(timestamp), fetch_files(conn))
    return delete_files_from_database(conn, map(os.path.normpath,
                                                chain(missing, modified)))

def store_html_sections(conn, html: str, sources: [str]):
    """Insert Html and Sections entries for html and sources.

    html
    : HTML body text.
    sources
    : List of source files from which the html was generated.
    """
    if not html.strip() or not sources:
        return
    cur = conn.cursor()
    cur.execute("PRAGMA foreign_keys=ON")
    cur.execute("INSERT INTO Html(body) VALUES (?)", (html,))
    lastrowid = cur.lastrowid
    query = "SELECT DISTINCT id FROM Notes WHERE filename IN ({})".format(
        ", ".join(map(utils.sqlite_string, sources)))
    insert = "INSERT OR IGNORE INTO Sections(note, html) VALUES (?, ?)"
    cur2 = conn.cursor()
    for row in cur.execute(query):
        cur2.execute(insert, (row[0], lastrowid))
    conn.commit()

def scan(conn, inputs, scan_options, convert_to_data_url):
    """Process inputs and store results in database.

    inputs shouldn't be an iterator, because it is used multiple times.
    """
    convert_to_data_url = "1" if convert_to_data_url else ""
    for batch in group_by_file_extension(inputs):
        files = list(batch)
        scan_input_list = " ".join(map(shlex.quote, files))
        with utils.make_temporary_file() as slipbox_sql,\
                utils.make_temporary_file(suffix=".html", text=True) as html:
            cmd = build_command(files, html, scan_options)
            out, err = utils.run_command(cmd, SLIPBOX_SQL=slipbox_sql,
                                         CONVERT_TO_DATA_URL=convert_to_data_url,
                                         SCAN_INPUT_LIST=scan_input_list)
            run_script_on_database(conn, slipbox_sql)

            if out:
                print(out.decode())
            if err:
                print(err.decode(), file=sys.stderr)

            contents = utils.get_contents(html)
        store_html_sections(conn, contents, inputs)



import setuptools

with open("README.md") as fp:
    long_description = fp.read()

setuptools.setup(
    name="slipbox",
    version="0.6.10",
    author="Levi Gruspe",
    author_email="mail.levig@gmail.com",
    description="Static-site generator for Zettelkasten notes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lggruspe/slipbox",
    packages=setuptools.find_packages(),
    package_data={
        "slipbox": ["filters/*.lua", "data/*.js", "*.sql"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Operating System :: POSIX",
    ],
    python_requires=">=3.8",
)



"""HTML elements."""

from textwrap import indent

class Elem:
    """Represents an HTML element."""
    def __init__(self, tag, *children, **attributes):
        self.tag = tag
        self.children = list(children)
        self.attributes = attributes.copy()

def render(elem, prefix=""):
    """Render HTML from Elem or str."""
    if isinstance(elem, str):
        return prefix + elem
    assert isinstance(elem, Elem)
    attrs = ""
    for key, val in elem.attributes.items():
        attrs += f" {key}={val!r}"
    children = indent("\n".join(render(child) for child in elem.children),
                      prefix + "  ")
    return "<{tag}{attrs}>\n{children}\n</{tag}>".format(
        tag=elem.tag,
        attrs=attrs,
        children=children)



"""Generate complete HTML containing all slipbox notes."""

import os
import shlex
import subprocess

from .templates import Elem, render
from .utils import make_temporary_file, write_lines

DUMMY_MARKDOWN = r"""$\,$
``` {.c style="display:none"}
```"""

def generate_active_htmls(conn):
    """Get HTML stored in the database for active sections."""
    sql = """
        SELECT body FROM Html WHERE id IN (SELECT html FROM Sections)
            ORDER BY id DESC
    """
    cur = conn.cursor()
    return map(lambda row: row[0], cur.execute(sql))

def generate_data(conn):
    """Generate slipbox data in javascript."""
    for nid, title in conn.execute("SELECT id, title FROM Notes ORDER BY id"):
        yield f"window.query.db.add(new Model.Note({nid}, {title!r}))"
    sql = "SELECT src, dest, annotation FROM ValidLinks"
    for src, dest, annotation in conn.execute(sql):
        yield f"""window.query.db.add(new Model.Link(
  window.query.note({src}), window.query.note({dest}), {annotation!r}))"""
    sql = "SELECT id, alias FROM ValidAliases ORDER BY id, alias"
    for nid, alias in conn.execute(sql):
        yield f"window.query.db.add(new Model.Alias({nid}, {alias!r}))"
    sql = "SELECT prev, next FROM Sequences ORDER BY prev, next"
    for prev, next_ in conn.execute(sql):
        yield f"window.query.db.add(new Model.Sequence({prev!r}, {next_!r}))"

def generate_javascript(conn):
    """Generate slipbox javascript code."""
    yield '<script type="module">'
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, "data/bundle.js")) as file:
        yield file.read().strip()
    yield '</script>'
    yield '<script type="text/javascript">'
    yield "window.addEventListener('DOMContentLoaded', () => {"
    yield from generate_data(conn)
    yield "window.initSlipbox()"
    yield "})"
    yield "</script>"

def create_bibliography(conn):
    """Create bibliography HTML section from database entries."""
    sql = "SELECT key, text FROM Bibliography ORDER BY key"
    items = []
    for key, text in conn.execute(sql):
        items.append(Elem("dt", Elem("a", f"[@{key[4:]}]", href='#' + key)))
        items.append(Elem("dd", text))
    section = Elem("section",
                   Elem("h1", "References"),
                   Elem("dl", *items),
                   id="references",
                   title="References",
                   **{"class": "level1"})
    return render(section)

def create_tags(conn):
    """Create HTML section that lists all tags."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    items = (Elem("li", Elem("a", tag, href=f"#{tag}")) for tag in tags)
    section = Elem("section",
                   Elem("h1", "Tags"),
                   Elem("ul", *items),
                   id="tags",
                   title="Tags",
                   **{"class": "level1"})
    return render(section)

def create_tag_page(conn, tag):
    """Create HTML section that lists all notes with the tag."""
    sql = """
        SELECT id, title FROM Tags NATURAL JOIN Notes WHERE tag = ? ORDER BY id
    """
    items = []
    for nid, title in conn.execute(sql, (tag,)):
        item = Elem("li", f"[{nid}] ", Elem("a", title, href=f"#{nid}"))
        items.append(item)
    section = Elem("section",
                   Elem("h1", Elem("a", tag, href="#tags", title="List of tags")),
                   Elem("ul", *items),
                   id=tag,
                   title=tag,
                   **{"class": "level1"})
    return render(section)

def create_tag_pages(conn):
    """Create all tag pages."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    return '\n'.join(create_tag_page(conn, tag) for tag in tags)

def create_reference_page(conn, reference):
    """Create HTML section that lists all notes that cite the reference."""
    sql = """
        SELECT note, title, text FROM Citations
            JOIN Notes ON Citations.note = Notes.id
                JOIN Bibliography ON Bibliography.key = Citations.reference
                    WHERE reference = ?
    """
    items = []
    text = ""
    for note, title, _text in conn.execute(sql, (reference,)):
        text = _text
        item = Elem("li", f"[{note}] ", Elem("a", title, href=f"#{note}"))
        items.append(item)
    section = Elem("section",
                   Elem("h1", Elem("a", '@' + reference[4:], href="#references")),
                   Elem("p", text),
                   Elem("ul", *items),
                   id=reference,
                   title=reference,
                   **{"class": "level1"})
    return render(section)

def create_reference_pages(conn):
    """Create all reference pages."""
    rows = conn.execute("SELECT key FROM Bibliography ORDER BY key")
    references = (row[0] for row in rows)
    return '\n'.join(create_reference_page(conn, ref) for ref in references)

def generate_complete_html(conn, options):
    """Create final HTML file with javascript."""
    with make_temporary_file(suffix=".md", text=True) as dummy,\
            make_temporary_file() as script,\
            make_temporary_file(suffix=".html", text=True) as html,\
            make_temporary_file(suffix=".html", text=True) as extra:
        write_lines(dummy, [DUMMY_MARKDOWN])
        write_lines(script, generate_javascript(conn))
        write_lines(html, generate_active_htmls(conn))
        with open(extra, "w") as file:
            print(create_tag_pages(conn), file=file)
            print(create_tags(conn), file=file)
            print(create_reference_pages(conn), file=file)
            print(create_bibliography(conn), file=file)
        cmd = "pandoc {dummy} -H {script} -B {html} -B {extra} --section-divs {options}".format(
            dummy=shlex.quote(dummy), script=shlex.quote(script),
            html=shlex.quote(html), options=options, extra=shlex.quote(extra))
        subprocess.run(shlex.split(cmd), check=False)
